from __future__ import annotations

import shutil
import time
import uuid
import warnings
from functools import lru_cache
from io import SEEK_END, SEEK_SET
from pathlib import Path
from tempfile import TemporaryFile
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
    overload,
)

from fused.core._realtime_ops import _parse_realtime_response

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd

import requests

import fused
import fused.models.request as request_models
from fused._auth import AUTHORIZATION
from fused._deserialize_parquet import parquet_to_df
from fused._global_api import set_api, set_api_class
from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._options import (
    DEV_DEFAULT_BASE_URL,
    PROD_DEFAULT_BASE_URL,
    STAGING_DEFAULT_BASE_URL,
)
from fused._options import options as OPTIONS
from fused._request import raise_for_status
from fused._str_utils import detect_passing_local_file_as_str
from fused.models.api import (
    JobConfig,
    JobStepConfig,
    ListDetails,
    UdfAccessToken,
    UdfAccessTokenList,
)
from fused.models.internal import Jobs, RunResponse
from fused.models.internal.job import CoerceableToJobId, _object_to_job_id
from fused.models.request import WHITELISTED_INSTANCE_TYPES
from fused.models.udf._udf_registry import UdfRegistry
from fused.models.udf.base_udf import METADATA_FUSED_EXPLORER_TAB
from fused.models.udf.udf import AnyBaseUdf, load_udf_from_response_data
from fused.warnings import (
    FusedIgnoredWarning,
    FusedNonProductionWarning,
    FusedOnPremWarning,
    FusedWarning,
)


def _detect_upload_length(data: Union[bytes, BinaryIO]) -> int:
    if hasattr(data, "tell") and hasattr(data, "seek"):
        # Looks like an IO, try to get size that way
        data.seek(0, SEEK_END)
        length = data.tell()
        # Reset to beginning
        data.seek(0, SEEK_SET)
        return length

    return len(data)


def resolve_udf_server_url(
    client_id: Optional[str] = None, *, base_url: Optional[str] = None
) -> str:
    if client_id and client_id.endswith("-staging"):
        return f"{base_url or STAGING_DEFAULT_BASE_URL}/realtime/{client_id}"

    if client_id is None:
        api = FusedAPI()
        client_id = api._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

    return f"{base_url or OPTIONS.base_url}/realtime/{client_id}"


class FusedAPI:
    """API for running jobs in the Fused service."""

    base_url: str

    def __init__(
        self,
        *,
        base_url: Optional[str] = None,
        set_global_api: bool = True,
        credentials_needed: bool = True,
    ):
        """Create a FusedAPI instance.

        Keyword Args:
            base_url: The Fused instance to send requests to. Defaults to `https://www.fused.io/server/v1`.
            set_global_api: Set this as the global API object. Defaults to True.
            credentials_needed: If True, automatically attempt to log in. Defaults to True.
        """
        if credentials_needed:
            AUTHORIZATION.initialize()
        base_url = base_url or OPTIONS.base_url

        self.base_url = base_url
        self._check_is_prod()

        if set_global_api:
            set_api(self)

    def _check_is_prod(self):
        if self.base_url in [STAGING_DEFAULT_BASE_URL, DEV_DEFAULT_BASE_URL]:
            warnings.warn(
                FusedNonProductionWarning(
                    "FusedAPI is connected to a development environment"
                )
            )
        elif self.base_url != PROD_DEFAULT_BASE_URL:
            warnings.warn(
                FusedOnPremWarning("FusedAPI is connected to an on-prem environment")
            )

    def download_table_bbox(
        self,
        path: str,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        n_rows: Optional[int] = None,
        columns: Optional[List[str]] = None,
        clip: bool = True,
        buffer: Optional[float] = None,
    ) -> Union["pd.DataFrame", "gpd.GeoDataFrame"]:
        url = f"{self.base_url}/table/download/bbox"

        params = request_models.GetTableBboxRequest(
            path=path,
            bbox_minx=minx,
            bbox_miny=miny,
            bbox_maxx=maxx,
            bbox_maxy=maxy,
            n_rows=n_rows,
            columns=columns,
            clip=clip,
            buffer=buffer,
        )

        self._check_is_prod()
        r = requests.get(
            url=url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return parquet_to_df(r.content)

    def start_job(
        self,
        config: Union[JobConfig, JobStepConfig],
        *,
        instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None,
        region: Optional[str] = None,
        disk_size_gb: Optional[int] = None,
        additional_env: Optional[Sequence[str]] = ("FUSED_CREDENTIAL_PROVIDER=ec2",),
        image_name: Optional[str] = None,
    ) -> RunResponse:
        """Execute an operation

        Args:
            config: the configuration object to run in the job.

        Keyword Args:
            instance_type: The AWS EC2 instance type to use for the job. Acceptable strings are "m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge", "r5.large", "r5.xlarge", "r5.2xlarge", "r5.4xlarge". Defaults to None.
            region: The AWS region in which to run. Defaults to None.
            disk_size_gb: The disk size to specify for the job. Defaults to None.
            additional_env: Any additional environment variables to be passed into the job, each in the form KEY=value. Defaults to None.
            image_name: Custom image name to run. Defaults to None for default image.
        """
        url = f"{self.base_url}/run"

        if isinstance(config, JobStepConfig):
            config = JobConfig(steps=[config])

        body = {"config": config.model_dump()}
        if additional_env:
            body["additional_env"] = additional_env
        if image_name:
            body["image_name"] = image_name

        params = request_models.StartJobRequest(
            region=region,
            instance_type=instance_type,
            disk_size_gb=disk_size_gb,
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def save_udf(
        self,
        udf: AnyBaseUdf,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        allow_public_read: bool = False,
        allow_public_list: bool = False,
    ) -> UdfRegistry:
        url = f"{self.base_url}/udf/by-id/{id}" if id else f"{self.base_url}/udf/new"
        body = request_models.SaveUdfRequest(
            slug=slug,
            udf_body=udf.model_dump_json(),
            udf_type=request_models.UdfType.auto,
            allow_public_read=allow_public_read,
            allow_public_list=allow_public_list,
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            json=body.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        # TODO: body
        return r.json()

    def delete_saved_udf(self, id: str):
        url = f"{self.base_url}/udf/by-id/{id}"

        self._check_is_prod()
        r = requests.delete(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        # TODO: body
        return r.json()

    def _get_udf(
        self,
        email_or_handle_or_id: str,
        slug: Optional[str] = None,
    ):
        if slug is None:  # first arg is id
            slug = email_or_handle_or_id
            email = self._whoami()["email"]
            url = f"{self.base_url}/udf/by-user-email/{email}/by-slug/{slug}"
        else:
            if "@" in email_or_handle_or_id:  # all valid email addresses contain @
                url = f"{self.base_url}/udf/by-user-email/{email_or_handle_or_id}/by-slug/{slug}"
            else:
                url = f"{self.base_url}/udf/by-user-handle/{email_or_handle_or_id}/by-slug/{slug}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_udf_by_id(self, id: str):
        url = f"{self.base_url}/udf/by-id/{id}"
        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_public_udf(
        self,
        id: str,
    ):
        url = f"{self.base_url}/udf/public/by-slug/{id}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_code_by_url(self, url: str):
        req_url = f"{self.base_url}/code-proxy/by-url"

        r = requests.get(
            url=req_url,
            params={
                "url": url,
            },
            headers=self._generate_headers(credentials_needed=False),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_udf_objects(
        self,
        n: int = 5,
        *,
        object_type: Literal["app", "udf"] = "udf",
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        by: Literal["name", "id", "slug"] = "name",
        whose: Literal["self", "public", "community", "team"] = "self",
    ):
        request_count = 0
        has_content = True
        udfs = []

        udf_type_includes: Dict[Literal["app", "udf"], List[str]] = {
            "app": ["app"],
            "udf": [
                "auto",
                "vector_tile",
                "vector_single",
                "raster",
                "raster_single",
            ],
        }[object_type]

        assert per_request >= 0

        while has_content:
            if whose == "self":
                url = f"{self.base_url}/udf/self"
            elif whose == "public" or whose == "community":
                url = f"{self.base_url}/udf/public"
            elif whose == "team":
                url = f"{self.base_url}/udf/exec-env/self"
            else:
                raise ValueError(
                    'Invalid value for `whose`, should be one of: "self", "public", "community", "team"'
                )

            params = request_models.ListUdfsRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            udfs_this_request = r.json()
            if udfs_this_request:
                non_app_udfs = list(
                    udf
                    for udf in udfs_this_request
                    if udf["udf_type"] in udf_type_includes
                )
                udfs.extend(non_app_udfs)
            else:
                has_content = False

            request_count += 1
            if len(udfs) >= n or (
                max_requests is not None and request_count == max_requests
            ):
                break

        deserialized_udfs = {}
        for udf in udfs:
            if "udf_body" in udf and udf["udf_body"]:
                try:
                    deserialized_udf = load_udf_from_response_data(udf)
                    udf_id = (
                        deserialized_udf.name
                        if by == "name"
                        else (
                            udf["id"]
                            if by == "id"
                            else (udf["slug"] if by == "slug" else None)
                        )
                    )

                    filtered_public_udf = (
                        whose == "public"
                        and deserialized_udf._get_metadata_safe(
                            METADATA_FUSED_EXPLORER_TAB
                        )
                        == "community"
                    ) or (
                        whose == "community"
                        and deserialized_udf._get_metadata_safe(
                            METADATA_FUSED_EXPLORER_TAB
                        )
                        != "community"
                    )

                    if udf_id is not None and not filtered_public_udf:
                        deserialized_udfs[udf_id] = deserialized_udf
                except Exception as e:
                    warnings.warn(
                        FusedWarning(
                            f"UDF {udf['slug']} ({udf['id']}) could not be deserialized: {e}"
                        ),
                    )
        return UdfRegistry(deserialized_udfs)

    def get_apps(
        self,
        n: int = 5,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        by: Literal["name", "id", "slug"] = "name",
        whose: Literal["self", "public", "community", "team"] = "self",
    ):
        return self._get_udf_objects(
            n=n,
            object_type="app",
            skip=skip,
            per_request=per_request,
            max_requests=max_requests,
            by=by,
            whose=whose,
        )

    def get_udfs(
        self,
        n: int = 5,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        by: Literal["name", "id", "slug"] = "name",
        whose: Literal["self", "public", "community", "team"] = "self",
    ):
        return self._get_udf_objects(
            n=n,
            object_type="udf",
            skip=skip,
            per_request=per_request,
            max_requests=max_requests,
            by=by,
            whose=whose,
        )

    def get_udf_access_tokens(
        self,
        n: Optional[int] = None,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        _whose: Literal["self", "all"] = "self",
    ) -> UdfAccessTokenList:
        request_count = 0
        has_content = True
        tokens = []

        assert per_request >= 0

        while has_content:
            url = f"{self.base_url}/udf-access-token/{_whose}"

            params = request_models.ListUdfAccessTokensRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            tokens_this_request = r.json()
            if tokens_this_request:
                tokens.extend(tokens_this_request)
            else:
                has_content = False

            request_count += 1
            if n is not None and (
                len(tokens) >= n
                or (max_requests is not None and request_count == max_requests)
            ):
                break

        tokens_deserialized = UdfAccessTokenList()
        for token in tokens:
            token_deserialized = UdfAccessToken.model_validate(token)
            token_deserialized.udf_email = self._whoami()["email"]
            tokens_deserialized.append(token_deserialized)

        return tokens_deserialized

    def get_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        token_obj = UdfAccessToken.model_validate_json(r.content)
        token_obj.udf_email = self._whoami()["email"]
        token_obj.udf_slug = self._get_udf_by_id(token_obj.udf_id)["slug"]
        return token_obj

    def delete_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        r = requests.delete(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        token_obj = UdfAccessToken.model_validate_json(r.content)
        token_obj.udf_email = self._whoami()["email"]
        token_obj.udf_slug = self._get_udf_by_id(token_obj.udf_id)["slug"]
        return token_obj

    def update_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
        *,
        client_id: Optional[str] = None,
        cache: Optional[bool] = None,
        metadata_json: Optional[Dict[str, Any]] = None,
        enabled: Optional[bool] = None,
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        body = request_models.UpdateUdfAccessTokenRequest(
            client_id=client_id,
            cache=cache,
            metadata_json=metadata_json,
            enabled=enabled,
        ).model_dump()  # type: ignore

        r = requests.post(
            url=url,
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        token_obj = UdfAccessToken.model_validate_json(r.content)
        token_obj.udf_email = self._whoami()["email"]
        token_obj.udf_slug = self._get_udf_by_id(token_obj.udf_id)["slug"]
        return token_obj

    def create_udf_access_token(
        self,
        udf_email_or_name_or_id: Optional[str] = None,
        /,
        udf_name: Optional[str] = None,
        *,
        udf_email: Optional[str] = None,
        udf_id: Optional[str] = None,
        client_id: Union[str, Ellipsis, None] = ...,
        cache: bool = True,
        metadata_json: Optional[Dict[str, Any]] = None,
        enabled: bool = True,
    ) -> UdfAccessToken:
        """
        Create a token for running a UDF. The token allows anyone who has it to run
        the UDF, with the parameters they choose. The UDF will run under your environment.

        The token does not allow running any other UDF on your account.

        Args:
            udf_email_or_name_or_id: A UDF ID, email address (for use with udf_name), or UDF name.
            udf_name: The name of the UDF to create the token for.

        Keyword Args:
            udf_email: The email of the user owning the UDF, or, if udf_name is None, the name of the UDF.
            udf_id: The backend ID of the UDF to create the token for.
            client_id: If specified, overrides which realtime environment to run the UDF under.
            cache: If True, UDF tiles will be cached.
            metadata_json: Additional metadata to serve as part of the tiles metadata.json.
            enable: If True, the token can be used.
        """
        if udf_id is not None:
            if (
                udf_name is not None
                or udf_email is not None
                or udf_email_or_name_or_id is not None
            ):
                warnings.warn(
                    FusedIgnoredWarning(
                        "All other ways of specifying the UDF are ignored in favor of udf_id."
                    ),
                )
                udf_name = None
                udf_email = None
        elif udf_name is not None:
            if udf_email_or_name_or_id is not None:
                if udf_email is not None:
                    warnings.warn(
                        FusedIgnoredWarning(
                            "All other ways of specifying the UDF are ignored in favor of the first argument and udf_name."
                        ),
                    )
                udf_email = udf_email_or_name_or_id
        elif udf_email_or_name_or_id is not None:
            if udf_name is not None:
                udf_email = udf_email_or_name_or_id
            else:
                # Need to figure out what exactly the first argument is and how it specifies a UDF
                is_valid_uuid = True
                try:
                    uuid.UUID(udf_email_or_name_or_id)
                except ValueError:
                    is_valid_uuid = False
                if is_valid_uuid:
                    udf_id = udf_email_or_name_or_id
                elif "/" in udf_email_or_name_or_id:
                    udf_email, udf_name = udf_email_or_name_or_id.split("/", maxsplit=1)
                else:
                    udf_name = udf_email_or_name_or_id
                    udf_email = self._whoami()["email"]
        else:
            raise ValueError("No UDF specified to create an access token for.")

        if client_id is Ellipsis:
            client_id = self._automatic_realtime_client_id()

        if client_id is Ellipsis:
            raise ValueError("Failed to detect realtime client ID")

        url = f"{self.base_url}/udf-access-token/new"

        metadata_json = metadata_json or {}

        body = request_models.CreateUdfAccessTokenRequest(
            udf_email=udf_email,
            udf_slug=udf_name,
            udf_id=udf_id,
            client_id=client_id,
            cache=cache,
            metadata_json=metadata_json,
            enabled=enabled,
        ).model_dump()  # type: ignore

        r = requests.post(
            url=url,
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        token_obj = UdfAccessToken.model_validate_json(r.content)
        token_obj.udf_email = self._whoami()["email"]
        token_obj.udf_slug = self._get_udf_by_id(token_obj.udf_id)["slug"]
        return token_obj

    def get_secret_value(
        self,
        key: str,
        client_id: Optional[str] = None,
    ) -> str:
        """Retrieve a secret value from the Fused service."""
        if client_id is None:
            client_id = self._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

        url = f"{resolve_udf_server_url(client_id=client_id, base_url=self.base_url)}/api/v1/secrets/{key}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def set_secret_value(
        self,
        key: str,
        value: str,
        client_id: Optional[str] = None,
    ) -> str:
        """Set a secret value on the Fused service."""
        if client_id is None:
            client_id = self._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

        body = request_models.UpdateSecretRequest(value=value).model_dump()
        url = f"{resolve_udf_server_url(client_id=client_id, base_url=self.base_url)}/api/v1/secrets/{key}"

        r = requests.put(
            url=url,
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def delete_secret_value(
        self,
        key: str,
        client_id: Optional[str] = None,
    ) -> str:
        """Delete a secret value on the Fused service."""
        if client_id is None:
            client_id = self._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

        url = f"{resolve_udf_server_url(client_id=client_id, base_url=self.base_url)}/api/v1/secrets/{key}"

        r = requests.delete(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def list_secrets(
        self,
        client_id: Optional[str] = None,
    ) -> Iterable[str]:
        """List available secret values on the Fused service.

        This may also be used to retrieve all secrets."""
        if client_id is None:
            client_id = self._automatic_realtime_client_id()

        if client_id is None:
            raise ValueError("Failed to detect realtime client ID")

        url = f"{resolve_udf_server_url(client_id=client_id, base_url=self.base_url)}/api/v1/secrets"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json().keys()

    def get_jobs(
        self,
        n: int = 5,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
    ) -> Jobs:
        """Get the job history.

        Args:
            n: The number of jobs to fetch. Defaults to 5.

        Keyword Args:
            skip: Where in the job history to begin. Defaults to 0, which retrieves the most recent job.
            per_request: Number of jobs per request to fetch. Defaults to 25.
            max_requests: Maximum number of requests to make. May be None to fetch all jobs. Defaults to 1.

        Returns:
            The job history.
        """
        request_count = 0
        has_content = True
        jobs = []
        original_skip = skip

        assert per_request >= 0

        while has_content:
            url = f"{self.base_url}/job/self"

            params = request_models.ListJobsRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            jobs_this_request = r.json()
            if jobs_this_request:
                jobs.extend(jobs_this_request)
            else:
                has_content = False

            request_count += 1
            if len(jobs) >= n or (
                max_requests is not None and request_count == max_requests
            ):
                break

        return Jobs(
            jobs=jobs[:n],
            n=n,
            skip=original_skip,
            per_request=per_request,
            max_requests=max_requests,
        )

    def get_job_config(self, job: CoerceableToJobId) -> JobConfig:
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/job/by-id/{job_id}/config"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=False,
        )
        raise_for_status(r)

        redirect_location = r.headers["location"]

        r2 = requests.get(
            url=redirect_location,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r2)

        return JobConfig.model_validate_json(r2.content)

    def get_status(self, job: CoerceableToJobId) -> RunResponse:
        """Fetch the status of a running job

        Args:
            job: the identifier of a job or a `RunResponse` object.

        Returns:
            The status of the given job.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/run/by-id/{job_id}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def get_logs(
        self,
        job: CoerceableToJobId,
        since_ms: Optional[int] = None,
    ) -> List[Any]:
        """Fetch logs for a job

        Args:
            job: the identifier of a job or a `RunResponse` object.
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

        Returns:
            Log messages for the given job.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/logs/{job_id}"
        params = {}
        if since_ms is not None:
            params["since_ms"] = since_ms
        r = requests.get(
            url=url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def tail_logs(
        self,
        job: CoerceableToJobId,
        refresh_seconds: float = 1,
        sample_logs: bool = True,
        timeout: Optional[float] = None,
        get_logs_retries: int = 1,
    ):
        """Continuously print logs for a job

        Args:
            job: the identifier of a job or a `RunResponse` object.
            refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
            sample_logs: if true, print out only a sample of logs. Defaults to True.
            timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
            get_logs_retries: Number of additional retries for log requests. Defaults to 1.
        """
        # TODO: Move this to the RunResponse object
        start_time = time.time()
        job = self.get_status(job)
        print(f"Logs for: {job.job_id}")

        def _tail_get_logs(
            job: RunResponse, since_ms: Optional[int] = None
        ) -> List[Any]:
            for _ in range(get_logs_retries + 1):
                try:
                    return self.get_logs(job, since_ms=since_ms)
                except requests.exceptions.RequestException:
                    # TODO: Don't use bare except
                    print("Server did not respond with logs")
                    # TODO: Backoff strategy
                    time.sleep(refresh_seconds)
            raise ValueError("Server did not respond with logs")

        r = _tail_get_logs(job)
        if job.status not in ["running", "pending"]:
            print(f"Job is not running ({job.status})")
            return

        if len(r) == 0:
            print("Configuring packages and waiting for logs...")
            while len(r) == 0:
                time.sleep(refresh_seconds)
                r = _tail_get_logs(job)
                if timeout is not None and time.time() - start_time > timeout:
                    raise TimeoutError("Timed out waiting for logs")

        last_message: Optional[str] = None
        last_since_ms: Optional[int] = None
        while True:
            time.sleep(refresh_seconds)
            r = _tail_get_logs(job, since_ms=last_since_ms)
            # If any results -- there may be none because we are filtering them with since_ms
            if len(r):
                current_message: str = r[-1]["message"]
                if last_message != current_message:
                    # If the most recent log line has changed, print it out
                    last_message = current_message
                    last_since_ms = r[-1]["timestamp"]

                    if sample_logs:
                        print(current_message.rstrip())
                    else:
                        for message in r:
                            print(message["message"].rstrip())

            if "ERROR" in current_message or self.get_status(job).status != "running":
                # Try to detect exit scenarios: an error has occured and the job will stop,
                # or the job is no longer in a running state.
                return

            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out")

    def wait_for_job(
        self,
        job: CoerceableToJobId,
        poll_interval_seconds: float = 5,
        timeout: Optional[float] = None,
    ) -> RunResponse:
        """Block the Python kernel until the given job has finished

        Args:
            job: the identifier of a job or a `RunResponse` object.
            poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
            timeout: The length of time in seconds to wait for the job. Defaults to None.

        Raises:
            TimeoutError: if waiting for the job timed out.

        Returns:
            The status of the given job.
        """
        # TODO: Move this to the RunResponse object
        start_time = time.time()
        status = self.get_status(job)
        while not status.terminal_status:
            time.sleep(poll_interval_seconds)
            status = self.get_status(job)
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out waiting for job")
        return status

    def cancel_job(self, job: CoerceableToJobId) -> RunResponse:
        """Cancel an existing job

        Args:
            job: the identifier of a job or a `RunResponse` object.

        Returns:
            A new job object.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/run/by-id/{job_id}/cancel"

        self._check_is_prod()
        r = requests.post(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def get_job_results(
        self, job: CoerceableToJobId, arg_index: int = -1
    ) -> Any | list[Any]:
        """Get the deserialized batch results of a job

        Args:
            job: the identifier of a job or a `RunResponse` object.
            arg_index: number representing the desired partition result of a batch job

        Returns:
            A single deserialized vector or raster, or a list of them
        """
        files = self._list_job_results(job)
        sorted_files = sorted(files, key=lambda o: int(o["name"].split(".")[0]))
        if arg_index > -1:
            file_obj = sorted_files[arg_index]
            return self._process_job_result_obj(file_obj)

        return [self._process_job_result_obj(fo) for fo in sorted_files]

    def _list_job_results(self, job: CoerceableToJobId) -> list[dict]:
        job_id = _object_to_job_id(job)
        results_url = f"{self.base_url}/run/by-id/{job_id}/results"
        r = requests.get(
            url=results_url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _process_job_result_obj(self, file_obj: dict) -> Any:
        r = requests.get(file_obj["url"])
        return _parse_realtime_response(r, r.headers["content-type"])

    def _whoami(self) -> Any:
        """
        Returns information on the currently logged in user
        """
        url = f"{self.base_url}/user/self"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _list_realtime_instances(self, *, whose: str = "self") -> List[Any]:
        """
        Returns information about available realtime instances
        """
        url = f"{self.base_url}/realtime-instance"
        if whose == "self":
            url += "/available"
        else:
            assert whose == "public", "whose must be 'public' or 'self'"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _automatic_realtime_client_id(self) -> Optional[str]:
        client_id = OPTIONS.realtime_client_id
        if client_id is None:
            instances = self._list_realtime_instances()
            if len(instances):
                instances = sorted(
                    instances,
                    key=lambda instance: instance.get("preference_rank", 0),
                    reverse=True,
                )
                client_id = instances[0]["client_id"]
            if OPTIONS.save_user_settings and client_id:
                OPTIONS.realtime_client_id = client_id

        return client_id

    @overload
    def list(self, path: str, *, details: Literal[True]) -> List[ListDetails]:
        ...

    @overload
    def list(self, path: str, *, details: Literal[False] = False) -> List[str]:
        ...

    def list(
        self, path: str, *, details: bool = False, client_id: Optional[str] = None
    ):
        list_request_url = f"{self.base_url}/files/list{'-details' if details else ''}"

        params = request_models.ListPathRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=list_request_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        result = r.json()
        if details:
            result = [ListDetails.model_validate(detail) for detail in result]
        return result

    def delete(
        self,
        path: str,
        max_deletion_depth: Union[int, Literal["unlimited"]] = 2,
        *,
        client_id: Optional[str] = None,
    ) -> bool:
        delete_request_url = f"{self.base_url}/files/delete"

        params = request_models.DeletePathRequest(
            path=path,
            max_deletion_depth=max_deletion_depth,
        ).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.delete(
            url=delete_request_url,
            params=params,
            json="{}",
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _resolve(
        self,
        path: str,
    ) -> List[str]:
        resolve_request_url = f"{self.base_url}/files/resolve"

        params = request_models.ResolvePathRequest(
            path=path,
        )

        r = requests.post(
            url=resolve_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def get(self, path: str, *, client_id: Optional[str] = None) -> bytes:
        get_request_url = f"{self.base_url}/files/get"

        params = request_models.GetPathRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=get_request_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=True,
        )
        raise_for_status(r)
        return r.content

    def download(
        self,
        path: str,
        local_path: Union[str, Path],
        *,
        client_id: Optional[str] = None,
    ) -> None:
        get_request_url = f"{self.base_url}/files/get"

        params = request_models.GetPathRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=get_request_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=True,
            stream=True,
        )
        raise_for_status(r)
        with open(local_path, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    def sign_url(self, path: str, *, client_id: Optional[str] = None) -> str:
        sign_request_url = f"{self.base_url}/files/sign"

        params = request_models.SignPathRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=sign_request_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def sign_url_prefix(
        self, path: str, *, client_id: Optional[str] = None
    ) -> Dict[str, str]:
        sign_prefix_request_url = f"{self.base_url}/files/sign_prefix"

        params = request_models.SignPathRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=sign_prefix_request_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def upload(self, path: str, data: Union[bytes, BinaryIO]) -> None:
        """Upload a binary blob to a cloud location"""
        return self._upload_signed(path=path, data=data)

    def _upload_tmp(self, extension: str, data: Union[bytes, BinaryIO]) -> str:
        """Upload a binary blob to a temporary cloud location, and return the new URL"""
        return self._upload_tmp_signed(extension=extension, data=data)

    def _upload_signed(
        self,
        path: str,
        data: Union[bytes, BinaryIO],
        *,
        client_id: Optional[str] = None,
    ) -> None:
        """Upload a binary blob to a cloud location"""
        upload_url = f"{self.base_url}/files/sign-upload"

        params = request_models.SignUploadRequest(
            path=path, content_length=_detect_upload_length(data)
        ).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=upload_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        signed_post = r.json()

        r2 = requests.post(
            url=signed_post["url"],
            data=signed_post["fields"],
            files={"file": data},
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r2)

    def _upload_tmp_signed(
        self,
        extension: str,
        data: Union[bytes, BinaryIO],
        *,
        client_id: Optional[str] = None,
    ) -> str:
        """Upload a binary blob to a temporary cloud location, and return the new URL"""
        upload_temp_url = f"{self.base_url}/files/sign-upload-temp"

        params = request_models.SignUploadTempRequest(
            extension=extension, content_length=_detect_upload_length(data)
        ).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.get(
            url=upload_temp_url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        sign_details = r.json()

        signed_post = sign_details["signed_post"]
        r2 = requests.post(
            url=signed_post["url"],
            data=signed_post["fields"],
            files={"file": data},
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r2)

        return sign_details["storage_url"]

    def _upload_direct(
        self,
        path: str,
        data: Union[bytes, BinaryIO],
        *,
        client_id: Optional[str] = None,
    ) -> None:
        """Upload a binary blob to a cloud location"""
        upload_url = f"{self.base_url}/files/upload"

        params = request_models.UploadRequest(path=path).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.put(
            url=upload_url,
            params=params,
            headers=self._generate_headers(
                {"Content-Type": "application/octet-stream"}
            ),
            data=data,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)

    def _upload_tmp_direct(
        self,
        extension: str,
        data: Union[bytes, BinaryIO],
        *,
        client_id: Optional[str] = None,
    ) -> str:
        """Upload a binary blob to a temporary cloud location, and return the new URL"""
        upload_temp_url = f"{self.base_url}/files/upload-temp"

        params = request_models.UploadTempRequest(extension=extension).model_dump()
        if client_id is None:
            client_id = self._automatic_realtime_client_id()
        if client_id:
            params["client_id"] = client_id

        r = requests.post(
            url=upload_temp_url,
            params=params,
            headers=self._generate_headers(
                {"Content-Type": "application/octet-stream"}
            ),
            data=data,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _upload_local_input(
        self, input: Union[str, Path, List[Union[str, Path]], "gpd.GeoDataFrame"]
    ) -> Union[str, List[str]]:
        """Upload local input (in-memory DataFrame or path to local file(s)).

        If uploaded, return a URL to it. Otherwise return input unchanged.
        """
        # in-memory (Geo)DataFrame
        if HAS_PANDAS and isinstance(input, PD_DATAFRAME):
            with TemporaryFile(dir=OPTIONS.temp_directory) as tmp:
                input.to_parquet(tmp)
                tmp.seek(0, SEEK_SET)
                input = self._upload_tmp(extension="parquet", data=tmp)
                return input

        # local file path(s)
        def _upload_path(input_path: Path) -> str:
            with open(input_path, "rb") as f:
                extension = input_path.name.rsplit(".", 1)[-1]
                uploaded_input = self._upload_tmp(extension=extension, data=f)
            return uploaded_input

        if isinstance(input, (str, Path)):
            input = detect_passing_local_file_as_str(input)
            if isinstance(input, Path):
                return _upload_path(input)
            else:
                # not a local file path, return as is
                return input
        elif isinstance(input, list):
            uploaded_inputs = []
            for input_path in input:
                input_path = detect_passing_local_file_as_str(input_path)
                if isinstance(input_path, Path):
                    uploaded_inputs.append(_upload_path(input_path))
                else:
                    uploaded_inputs.append(input_path)
            return uploaded_inputs
        else:
            raise TypeError(f"Unsupported input type: {type(input)}")

    def _health(self) -> bool:
        """Check the health of the API backend"""
        r = requests.get(f"{self.base_url}/health", timeout=OPTIONS.request_timeout)
        raise_for_status(r)
        return True

    def auth_token(self) -> str:
        """
        Returns the current user's Fused environment (team) auth token
        """
        url = f"{self.base_url}/execution-env/token"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _generate_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        *,
        credentials_needed: bool = True,
    ) -> Dict[str, str]:
        if headers is None:
            headers = {}

        common_headers = {
            "Fused-Py-Version": fused.__version__,
            **headers,
        }

        if AUTHORIZATION.is_configured() or credentials_needed:
            common_headers[
                "Authorization"
            ] = f"{AUTHORIZATION.credentials.auth_scheme} {AUTHORIZATION.credentials.access_token}"

        return common_headers

    @lru_cache
    def dependency_whitelist(self) -> str:
        sign_request_url = f"{self.base_url}/internal/dependency-whitelist"
        r = requests.get(
            url=sign_request_url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()


set_api_class(FusedAPI)



import base64
from io import BytesIO
from typing import TYPE_CHECKING, Iterable, Optional, Union

if TYPE_CHECKING:
    import geopandas as gpd


def _normalize_url(url: str) -> str:
    if url.endswith("/"):
        return url[:-1]
    return url


def get_chunks_metadata(url: str) -> "gpd.GeoDataFrame":
    """Returns a GeoDataFrame with each chunk in the table as a row.

    Args:
        url: URL of the table.
    """
    import geopandas as gpd
    import pandas as pd
    import pyarrow.parquet as pq
    import shapely

    url = _normalize_url(url) + "/_sample"

    with pq.ParquetFile(url) as file:
        # do not use pq.read_metadata as it may segfault in versions >= 12 (tested on 15.0.1)
        sample_metadata = file.metadata

    if (
        b"fused:format_version" not in sample_metadata.metadata.keys()
        or sample_metadata.metadata[b"fused:format_version"] != b"5"
    ):
        raise ValueError(
            "Dataset does not have Fused metadata or it is an incompatible version."
        )

    metadata_bytes = sample_metadata.metadata[b"fused:_metadata"]
    metadata_bytes = base64.decodebytes(metadata_bytes)
    metadata_bio = BytesIO(metadata_bytes)
    df = pd.read_parquet(metadata_bio)
    geoms = shapely.box(
        df["bbox_minx"], df["bbox_miny"], df["bbox_maxx"], df["bbox_maxy"]
    )
    return gpd.GeoDataFrame(df, geometry=geoms, crs="EPSG:4326")


def get_chunk_from_table(
    url: str,
    file_id: Union[str, int, None],
    chunk_id: Optional[int],
    *,
    columns: Optional[Iterable[str]] = None,
) -> "gpd.GeoDataFrame":
    """Returns a chunk from a table and chunk coordinates.

    This can be called with file_id and chunk_id from `get_chunks_metadata`.

    Args:
        url: URL of the table.
        file_id: File ID to read.
        chunk_id: Chunk ID to read.
    """
    import geopandas as gpd
    import geopandas.io.arrow
    import pyarrow.parquet as pq

    if file_id is None:
        data = gpd.read_parquet(url)
    else:
        url = _normalize_url(url) + f"/{file_id}.parquet"

        with pq.ParquetFile(url) as file:
            if chunk_id is not None:
                table = file.read_row_group(chunk_id, columns=columns)
            else:
                table = file.read(columns=columns)

            if b"geo" in table.schema.metadata:
                data = geopandas.io.arrow._arrow_to_geopandas(table)
            else:
                data = table.to_pandas()

    if isinstance(data, (gpd.GeoDataFrame, gpd.GeoSeries)):
        if data.crs is None:
            data = data.set_crs(4326)
    return data



from typing import Iterable, Optional, Union

import geopandas as gpd

from ._impl._table_ops_impl import get_chunk_from_table as _get_chunk_from_table
from ._impl._table_ops_impl import get_chunks_metadata as _get_chunks_metadata


def get_chunks_metadata(url: str) -> gpd.GeoDataFrame:
    """Returns a GeoDataFrame with each chunk in the table as a row.

    Args:
        url: URL of the table.
    """
    return _get_chunks_metadata(url)


def get_chunk_from_table(
    url: str,
    file_id: Union[str, int, None],
    chunk_id: Optional[int],
    *,
    columns: Optional[Iterable[str]] = None,
) -> gpd.GeoDataFrame:
    """Returns a chunk from a table and chunk coordinates.

    This can be called with file_id and chunk_id from `get_chunks_metadata`.

    Args:
        url: URL of the table.
        file_id: File ID to read.
        chunk_id: Chunk ID to read.

    Keyword Args:
        columns: Read only the specified columns.
    """
    return _get_chunk_from_table(
        url=url,
        file_id=file_id,
        chunk_id=chunk_id,
        columns=columns,
    )



from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union

from pydantic import BaseModel

from fused_batch._options import options as OPTIONS

SAMPLE_KEY_FIELDS = frozenset(("input", "input_left", "input_right", "buffer_distance"))
_sample_cache: Dict[Tuple[str, Optional[str], Optional[int], Optional[int]], Any] = {}

SampleType = TypeVar("SampleType")


def size() -> int:
    return len(_sample_cache)


def clear() -> None:
    """Clear the cache contents"""
    _sample_cache.clear()


def get_or_cache_sample(
    job_config: BaseModel,
    sample_coords: Tuple[Union[str, int, None], Optional[int], Optional[int]],
    generator: Callable[[], SampleType],
) -> SampleType:
    if not OPTIONS.cache.enable:
        return generator()

    job_json = job_config.json(include=SAMPLE_KEY_FIELDS)

    if sample_coords is not None:
        # Ensure the file ID is a string
        file_id = (
            str(sample_coords[0]) if sample_coords[0] is not None else sample_coords[0]
        )
        key = (job_json, file_id, sample_coords[1], sample_coords[2])
    else:
        key = (job_json, None, None, None)

    if key in _sample_cache:
        return _sample_cache[key]
    else:
        sample = generator()
        # TODO: Support a maximum size of the cache
        # TODO: Support an on disk cache instead of in memory
        _sample_cache[key] = sample
        return sample



from pathlib import Path
from typing import Protocol


class ExecutionContextProtocol(Protocol):
    @property
    def partition_tempdir(self) -> Path:
        """A partition-level temporary directory for user use during the job.

        A new directory is provided for each file."""
        ...

    @property
    def tempdir(self) -> Path:
        """A chunk-level temporary directory for user use during the job.

        A new directory is provided for each chunk."""
        ...


class LocalContext(ExecutionContextProtocol):
    @property
    def partition_tempdir(self) -> Path:
        raise NotImplementedError()

    @property
    def tempdir(self) -> Path:
        raise NotImplementedError()


context: ExecutionContextProtocol = LocalContext()
local_context = LocalContext()



from typing import Union

from pydantic import FileUrl, UrlConstraints
from pydantic_core import SchemaValidator, Url, core_schema
from typing_extensions import Annotated

S3Url = Annotated[Url, UrlConstraints(allowed_schemes=["s3"])]
GCSUrl = Annotated[Url, UrlConstraints(allowed_schemes=["gs"])]
FusedIntermediaryUrl = Annotated[
    Url, UrlConstraints(allowed_schemes=["fused_intermediary"])
]
FusedTeamUrl = Annotated[Url, UrlConstraints(allowed_schemes=["fd"])]


DatasetUrl = Union[S3Url, GCSUrl, FileUrl, FusedIntermediaryUrl, FusedTeamUrl]

dataset_url_schema = core_schema.url_schema()
dataset_url_schema_validator = SchemaValidator(dataset_url_schema)



from __future__ import annotations

import json
import warnings
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Optional

from fused._environment import infer_display_method
from fused._options import options as OPTIONS
from fused.warnings import FusedWarning

if TYPE_CHECKING:
    import pandas as pd


def deserialize_tiff(content: bytes):
    import rasterio
    import rioxarray
    import xarray as xr

    with NamedTemporaryFile(
        dir=OPTIONS.temp_directory, prefix="udf_result", suffix=".tiff"
    ) as ntmp:
        with open(ntmp.name, "wb") as f:
            f.write(content)
        with rasterio.open(ntmp.name) as dataset:
            rda = rioxarray.open_rasterio(
                dataset,
                masked=True,
            )
            metadata = dataset.tags(ns="fused")
        orig_type = metadata.get("orig_type")
        if orig_type == "numpy.ndarray":
            data = rda.values
            if len(json.loads(metadata["shape"])) < rda.ndim:
                data = data.squeeze()
            if "bounds" in metadata:
                bounds = json.loads(metadata["bounds"])
                data = (data, bounds)
        elif orig_type == "xarray.DataArray":
            data = rda
        else:
            key = metadata.get("key") or "image"
            data = xr.Dataset({key: rda})
    return data


def deserialize_png(content: bytes):
    import xarray as xr
    from PIL import Image

    display_method = infer_display_method(None, None)
    if display_method.show_widget:
        from IPython.display import Image as IPythonImage
        from IPython.display import display

        display(IPythonImage(data=content, format="png"))

    image = Image.open(BytesIO(content))
    width, height = image.size
    if len(image.getbands()) == 1:
        image_data = list(image.getdata())
        image_data = [
            image_data[i : i + width] for i in range(0, len(image_data), width)
        ]
        data_array = xr.DataArray(image_data, dims=["y", "x"])
    else:
        image_data = []
        for band in range(len(image.getbands())):
            band_data = list(image.getdata(band=band))
            band_data = [
                band_data[i : i + width] for i in range(0, len(band_data), width)
            ]
            image_data.append(band_data)
        data_array = xr.DataArray(image_data, dims=["band", "y", "x"])

    # Create the dataset with image, latitude, and longitude data
    dataset = xr.Dataset({"image": data_array})

    return dataset


def deserialize_parquet(content: bytes):
    import pyarrow.parquet as pq

    output_df: Optional[pd.DataFrame] = None

    res_buf = BytesIO(content)

    meta = pq.read_metadata(res_buf)
    if b"geo" in meta.metadata:
        try:
            import geopandas as gpd

            output_df = gpd.read_parquet(res_buf)
        except ValueError as e:
            warnings.warn(
                f"Result has geo metadata but could not be loaded in GeoPandas: {e}",
                FusedWarning,
            )
    if output_df is None:
        import pandas as pd

        output_df = pd.read_parquet(res_buf)

    # Force all column names to be strings
    output_df.columns = [str(x) for x in output_df.columns]

    return output_df



# ruff: noqa: F401

from .dataset import (
    AnyDatasetInput,
    AnyDatasetOutput,
    DatasetInput,
    DatasetInputBase,
    DatasetInputType,
    DatasetInputV2,
    DatasetInputV2Table,
    DatasetInputV2Type,
    DatasetOutput,
    DatasetOutputBase,
    DatasetOutputType,
    DatasetOutputV2,
    SampleStrategy,
)
from .job import JobResponse, Jobs, RunResponse



from typing import Optional

from fused_batch._options import options as OPTIONS
from fused_batch.models.api import Table
from fused_batch.models.api.job import _get_chunk


def post_open_table(
    table: Table,
    fetch_samples: Optional[bool] = None,
):
    should_fetch_samples = False
    if fetch_samples is not None:
        should_fetch_samples = fetch_samples
    elif OPTIONS.open.fetch_samples is not None:
        should_fetch_samples = OPTIONS.open.fetch_samples

    if should_fetch_samples:
        try:
            # TODO: refactor this
            base_path, table_name = table.url.rstrip("/").rsplit("/", maxsplit=1)
            first_chunk = table.chunk_metadata[0]
            table.sample = _get_chunk(
                base_path=base_path,
                tables=[table_name],
                file_id=first_chunk.file_id,
                chunk_id=first_chunk.chunk_id,
                n_rows=10,
            )
        except:  # noqa E722
            table.sample = None



from typing import TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


def _maybe_inplace(obj: T, inplace: bool) -> T:
    if inplace:
        return obj
    else:
        return obj.model_copy(deep=True)



from __future__ import annotations

import warnings
from enum import Enum
from typing import Dict, List, Literal, Optional, Union
from urllib.parse import urlparse

from pydantic import BaseModel, Field, RootModel, StrictBool, StrictStr, field_validator
from typing_extensions import Annotated

from fused_batch._constants import DEFAULT_TABLE_NAMES
from fused_batch._str_utils import append_url_part
from fused_batch.warnings import FusedDeprecationWarning

from .._project_aware import FusedProjectAware
from ..base import FusedBaseModel
from ..urls import dataset_url_schema_validator


class DatasetInputBase(FusedBaseModel):
    type: Literal[None, "v2"]


class DatasetInput(DatasetInputBase):
    """(Deprecated) A class to describe a dataset to be used in a join or map operation"""

    type: Literal[None] = None

    base_path: StrictStr
    """The base path of the input dataset."""
    tables: List[StrictStr] = Field(default_factory=lambda: list(DEFAULT_TABLE_NAMES))
    """The list of table names to fetch in the operation."""

    cache_locally: StrictBool = False
    """Whether to cache all files from the input locally."""

    read_sidecar_files: Optional[Dict[StrictStr, StrictBool]] = None
    """Whether to read sidecar files from the input as part of the operation."""

    first_n: Optional[int] = None
    """Process only the first N files of the dataset. This is for debugging and benchmarking purposes.

    Note that this can produce output tables that are missing files.
    """

    file_sort_field: Optional[str] = None
    """When processing files, use this field on the metadata table to sort the files"""

    @field_validator("base_path")
    @classmethod
    def validate_base_path(cls, val):
        dataset_url_schema_validator.validate_python(val)
        return val

    def __init__(self, **data):
        super().__init__(**data)
        warnings.warn("DatasetInput class is deprecated", FusedDeprecationWarning)


class DatasetInputType(str, Enum):
    V2 = "v2"
    """Zip or Union operation"""


class DatasetInputV2Type(str, Enum):
    ZIP = "zip"
    """Append column-wise within partitions"""
    UNION = "union"
    """Append partitions"""


class DatasetInputV2Table(BaseModel):
    url: StrictStr
    read_sidecar_files: Optional[StrictBool] = None
    cache_locally: StrictBool = False

    @field_validator("url")
    @classmethod
    def validate_base_path(cls, val):
        dataset_url_schema_validator.validate_python(val)
        return val


class DatasetInputV2(DatasetInputBase, FusedProjectAware):
    type: Literal["v2"] = "v2"

    operation: DatasetInputV2Type

    tables: List[DatasetInputV2Table]
    # TODO: columns: Any = None
    # TODO: rename columns
    # TODO: Move buffer_distance here

    first_n: Optional[int] = None
    """Process only the first N files of the dataset. This is for debugging and benchmarking purposes.

    Note that this can produce output tables that are missing files."""
    file_sort_field: Optional[str] = None
    """When processing files, use this field on the metadata table to sort the files"""

    @classmethod
    def from_table_url(cls, url: str, *, cache_locally: bool = False) -> DatasetInputV2:
        """Create a DatasetInputV2 that reads a single table from a URL."""
        return cls(
            # The specific operation isn't important, ZIP is just a convenient default.
            operation=DatasetInputV2Type.ZIP,
            tables=[DatasetInputV2Table(url=url, cache_locally=cache_locally)],
        )


AnyDatasetInput = Annotated[
    Union[
        DatasetInput,
        DatasetInputV2,
    ],
    Field(..., discriminator="type"),
]


class RootAnyDatasetInput(RootModel[AnyDatasetInput]):
    pass


class SampleStrategy(str, Enum):
    """How to generate output samples"""

    EMPTY = "empty"
    """Do not generate a sample"""
    FIRST_CHUNK = "first_chunk"
    """The sample is from the first chunk"""
    GEO = "geo"
    """Geographically sample"""


class DatasetOutputType(str, Enum):
    V2 = "v2"
    """Save as a table to a URL"""


class DatasetOutputBase(BaseModel):
    type: Literal[None, "v2"]

    save_index: Optional[StrictBool] = None
    """Whether to override saving the output index."""

    sample_strategy: Optional[SampleStrategy] = None
    """How to generate output samples, or None for the default."""

    overwrite: bool = False
    """Whether the API should overwrite the output dataset if it already exists."""

    def from_str(
        s: Optional[str], project_url: Optional[str] = None
    ) -> AnyDatasetOutput:
        try:
            parsed = urlparse(s)
            if parsed.scheme:
                output = DatasetOutputV2(url=s)
                output._project_url = project_url
                return output
        except (ValueError, TypeError, AttributeError):
            pass

        if project_url is not None:
            url = append_url_part(project_url, s) if s is not None else None
            output = DatasetOutputV2(url=url)
            output._project_url = project_url
            return output

        if s is None:
            return DatasetOutputV2()

        # Parsing as URL failed
        return DatasetOutput(table=s)


class DatasetOutput(DatasetOutputBase):
    """(Deprecated) Output that writes a table to a dataset"""

    type: Literal[None] = None
    base_path: Optional[StrictStr] = None
    table: Optional[StrictStr] = None

    def __init__(self, **data):
        super().__init__(**data)
        warnings.warn("DatasetOutput class is deprecated", FusedDeprecationWarning)


class DatasetOutputV2(DatasetOutputBase, FusedProjectAware):
    """Output that writes a table to a URL"""

    type: Literal["v2"] = "v2"

    url: Optional[StrictStr] = None
    """Table URL to write to"""

    def to_v1(self) -> DatasetOutput:
        """Converts this output object to a DatasetOutput (V1)"""
        if self.url is None:
            return DatasetOutput(
                base_path=None,
                table=None,
                save_index=self.save_index,
                sample_strategy=self.sample_strategy,
            )

        base_path, table = self.url.rstrip("/").rsplit("/", maxsplit=1)
        return DatasetOutput(
            base_path=base_path,
            table=table,
            save_index=self.save_index,
            sample_strategy=self.sample_strategy,
            overwrite=self.overwrite,
        )

    @property
    def table(self) -> Optional[str]:
        """Returns the table name for this output"""
        if self.url:
            return self.url.rstrip("/").rsplit("/", maxsplit=1)[1]
        return None


AnyDatasetOutput = Annotated[
    Union[
        DatasetOutput,
        DatasetOutputV2,
    ],
    Field(..., discriminator="type"),
]



# TODO: This file is no longer the most recent -- use fused.core.run_* instead
# This file is only for running non-saved (code included) UDFs
from __future__ import annotations

import json
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pyarrow as pa
import requests
from loguru import logger

from fused._options import options as OPTIONS
from fused.api.api import FusedAPI, resolve_udf_server_url
from fused.core._realtime_ops import serialize_realtime_params
from fused.core._serialization import (
    deserialize_parquet,
    deserialize_png,
    deserialize_tiff,
)
from fused.models import AnyJobStepConfig
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.models.udf.udf import AnyBaseUdf

from ..core._impl._realtime_ops_impl import get_recursion_factor

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd


def run_tile(
    x: float,
    y: float,
    z: float,
    data: Optional[
        Union[pd.DataFrame, gpd.GeoDataFrame, pa.Table, str, Path, Any]
    ] = None,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    cache_max_age: Optional[int] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers({"Content-Type": "application/json"})
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": data,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    r = requests.post(
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=headers,
        timeout=OPTIONS.request_timeout,
    )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def run(
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    read_options: Optional[Dict] = None,
    client_id: Optional[str] = None,
    cache_max_age: Optional[int] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
        read_options: If not None, options for reading `df` that will be passed to GeoPandas.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    method = "POST"
    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers({"Content-Type": "application/json"})
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    r = requests.request(
        method=method,
        url=url,
        params=req_params,
        json=post_attr_json,
        headers=post_attr_headers,
        timeout=OPTIONS.request_timeout,
    )
    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return _process_response(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_tile_async(
    x: float,
    y: float,
    z: float,
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    cache_max_age: Optional[int] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> UdfEvaluationResult:  # TODO: return png
    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    url = f"{udf_server_url}/api/v1/run/udf/tiles/{z}/{x}/{y}"

    # Headers
    recursion_factor = get_recursion_factor()
    headers = api._generate_headers({"Content-Type": "application/json"})
    headers["Fused-Recursion"] = f"{recursion_factor}"

    # Payload
    post_attr_json = {
        "data_left": None,
        "data_right": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Params
    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=headers,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


async def run_async(
    step_config: Optional[AnyJobStepConfig] = None,
    params: Optional[Dict[str, str]] = None,
    *,
    print_time: bool = False,
    client_id: Optional[str] = None,
    cache_max_age: Optional[int] = None,
    dtype_out_vector: str = "parquet",
    dtype_out_raster: str = "tiff",
) -> pd.DataFrame:
    """Run a UDF over a DataFrame.

    Args:
        step_config: AnyJobStepConfig.
        params: Additional parameters to pass to the UDF. Must be JSON serializable.

    Keyword Args:
        print_time: If True, print the amount of time taken in the request.
    """
    # TODO: This function is too complicated

    time_start = time.perf_counter()
    # We need to get the auth headers from the FusedAPI. Don't enable set_global_api
    # to avoid messing up the user's environment.
    api = FusedAPI(set_global_api=False)

    udf_server_url = resolve_udf_server_url(client_id)

    assert step_config is not None
    # Apply parameters, if we want to step_config that's returned to have this,
    # overwrite step_config.
    step_config_with_params = step_config.set_udf(
        udf=step_config.udf, parameters=serialize_realtime_params(params)
    )

    # Note: Custom UDF uses the json POST attribute.
    url = f"{udf_server_url}/api/v1/run/udf"

    # This is the body for when step_config_with_params.type == "udf".
    body = {
        "data_left": None,
        "step_config": step_config_with_params.model_dump_json(),
        "dtype_in": "json",
        # TODO remove dtype_out
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }

    post_attr_json = body

    recursion_factor = get_recursion_factor()
    post_attr_headers = api._generate_headers({"Content-Type": "application/json"})
    post_attr_headers["Fused-Recursion"] = f"{recursion_factor}"

    req_params = {
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": dtype_out_vector,
        "dtype_out_raster": dtype_out_raster,
    }
    # Make request
    start = time.time()

    if OPTIONS.pyodide_async_requests:
        import pyodide.http
        import yarl

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            method="POST",
            headers=post_attr_headers,
            body=json.dumps(post_attr_json),
            # TODO: timeout
        )
    else:
        import aiohttp

        async with aiohttp.ClientSession() as session:
            r = await session.post(
                url=url,
                params=req_params,
                json=post_attr_json,
                headers=post_attr_headers,
                # TODO: timeout
            )

    end = time.time()
    if print_time:
        logger.info(f"Time in request: {end - start}")

    time_end = time.perf_counter()
    time_taken_seconds = time_end - time_start

    return await _process_response_async(
        r, step_config=step_config_with_params, time_taken_seconds=time_taken_seconds
    )


def _process_response(
    r: requests.Response,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: Optional[bytes] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    try:
        result_content = r.content
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        has_exception = fused_metadata.get("has_exception", False)

        # Extract udf.
        udf: Optional[AnyBaseUdf] = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status_code == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                data = None

            # Else, process response output.
            elif r.headers["content-type"] == "image/png":
                data = deserialize_png(result_content)

            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                data = deserialize_tiff(r.content)

            else:
                # assume parquet
                data = deserialize_parquet(result_content)

            return UdfEvaluationResult(
                data=data,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
            )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(r.text)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = r.text

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise


async def _process_response_async(
    r,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: Optional[bytes] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    try:
        if r.status != 200 and r.status != 422:
            raise ValueError(await r.read())

        result_content = (
            await r.read() if not OPTIONS.pyodide_async_requests else await r.bytes()
        )
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        has_exception = fused_metadata.get("has_exception", False)

        # Extract udf.
        udf: Optional[AnyBaseUdf] = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                data = None

            # Else, process response output.
            elif r.headers["content-type"] == "image/png":
                data = deserialize_png(result_content)

            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                data = deserialize_tiff(r.content)

            else:
                # assume parquet
                data = deserialize_parquet(result_content)

            return UdfEvaluationResult(
                data=data,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
            )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(result_content)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = result_content.decode()

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message or "Unknown error occured",
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import uuid
from functools import partial
from html import escape
from typing import TYPE_CHECKING, Any, Callable, List, Optional

from fused._formatter.common import copyable_text, icon, load_static_files
from fused._options import options as OPTIONS

if TYPE_CHECKING:
    from fused.models.internal import JobResponse, Jobs, RunResponse


def collapsible_section(
    name,
    inline_details: str = "",
    details: str = "",
    n_items: Optional[int] = None,
    enabled: bool = True,
    collapsed: bool = False,
):
    # "unique" id to expand/collapse the section
    data_id = "section-" + str(uuid.uuid4())

    has_items = n_items is not None and n_items
    n_items_span = "" if n_items is None else f" <span>({n_items})</span>"
    enabled_str = "" if enabled and has_items else "disabled"
    collapsed_str = "" if collapsed or not has_items else "checked"
    tip = " title='Expand/collapse section'" if enabled else ""

    return (
        f"<input id='{data_id}' class='xr-section-summary-in' "
        f"type='checkbox' {enabled_str} {collapsed_str}>"
        f"<label for='{data_id}' class='xr-section-summary' {tip}>"
        f"{name}:{n_items_span}</label>"
        f"<div class='xr-section-inline-details'>{inline_details}</div>"
        f"<div class='xr-section-details'>{details}</div>"
    )


def _mapping_section(
    mapping,
    name: str,
    details_func: Callable[[Any], str],
    max_items_collapse,
    enabled: bool = True,
):
    n_items = len(mapping)
    expanded = n_items < max_items_collapse
    collapsed = not expanded

    return collapsible_section(
        name,
        details=details_func(mapping),
        n_items=n_items,
        enabled=enabled,
        collapsed=collapsed,
    )


def _obj_repr(obj, header_components, sections, sections_class="xr-sections"):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='{sections_class}'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def render_command_toolkit(job_id: str):
    """Render a command toolkit for a job"""
    text = """
    <h4 style="margin-left: 20px;">Command Toolkit</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    """
    for command_description, command_code in [
        # (
        #     "Get job config",
        #     f'fused.experimental.job("{job_id}", content_type="fused_job_id")',
        # ),
        # ("Get status", f'fused.api.job_get_status("{job_id}")'),
        ("Tail logs", f'fused.api.job_tail_logs("{job_id}")'),
        ("Print logs", f'fused.api.job_print_logs("{job_id}")'),
        ("Get execution time", f'fused.api.job_get_exec_time("{job_id}")'),
        ("Cancel", f'fused.api.job_cancel("{job_id}")'),
    ]:
        text += f"<li>{copyable_text(command_code, show_text=False)} {command_description} <code>{command_code}</code></li>"
    text += "</ul>"
    return text


def summarize_job(
    job: JobResponse,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = f"{job.job_status}" if job.job_status else "?"
    name = escape(str(job.id))
    dtype = ""  # f"rows:{table.num_rows}"

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    # data_id = "data-" + str(uuid.uuid4())
    # chunk_repr_disabled = "disabled" if table.sample is None else ""

    # Convert DataFrame to HTML table
    # TODO: escape html characters # escape(inline_variable_array_repr(variable, 35))
    preview = f"{job.creation_date}"

    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Details</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    <li>ID: {copyable_text(job.id)}</li>
    <li>Instance: {copyable_text(job.instance_id) if job.instance_id else "<code>None</code>"}</li>
    <li>Instance type: {copyable_text(job.instance_type) if job.instance_type else "<code>None</code>"}</li>
    <li>Job status: <code>{job.job_status}</code> {f"(last updated {job.job_status_date})" if job.job_status_date else ""}</li>
    <li><a href="{OPTIONS.base_web_url}/job_status/{job.id}">Logs</a></li>
    </ul>

    {render_command_toolkit(job.id)}
    """

    attrs_icon = icon("icon-file-text2")

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<label></label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
    )


def summarize_jobs(jobs: List[JobResponse]):
    li_items = []
    for job in jobs:
        li_content = summarize_job(job, is_index=False)
        li_items.append(f"<li class='xr-var-item'>{li_content}</li>")

    vars_li = "".join(li_items)

    return f"<ul class='xr-var-list'>{vars_li}</ul>"


job_section = partial(
    _mapping_section,
    name="Jobs",
    details_func=summarize_jobs,
    max_items_collapse=25,
)


def fused_jobs_repr(jobs: Jobs) -> str:
    obj_type = f"fused.{jobs.__class__.__name__}"
    header_components = [
        f"<div class='xr-obj-type'>{escape(obj_type)}</div>",
        # get_object_settings(join, exclude=["input_left", "input_right", "udf", "type"]),
    ]

    sections: List[str] = [
        job_section(jobs.jobs),
    ]

    return _obj_repr(jobs, header_components, sections)


def fused_runresponse_repr(runresponse: RunResponse) -> str:
    terminal = " - final status" if runresponse.terminal_status else ""
    icons_svg, css_style = load_static_files()
    return f"""
    <div>
    {icons_svg}<style>{css_style}</style>
    <h4 style="margin-left: 20px;">Job Run</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    <li>Job ID: {copyable_text(runresponse.job_id)}</li>
    <li>Job name: {copyable_text(runresponse.job_name)}</li>
    <li>Instance ID: {copyable_text(runresponse.instance_id)} [{copyable_text(runresponse.instance_type)}] ({runresponse.status}{terminal})</li>
    <li><a href="{OPTIONS.base_web_url}/job_status/{runresponse.job_id}">Logs</a></li>
    </ul>

    {render_command_toolkit(runresponse.job_id)}
    </div>
    """



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict

from fused_batch._global_api import get_api

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

if TYPE_CHECKING:
    from fused_batch.api import FusedAPI

# Pydantic methods to remove in __dir__
PYDANTIC_METHODS = {
    "Config",
    "construct",
    "copy",
    "from_orm",
    "json",
    "parse_file",
    "parse_obj",
    "schema",
    "schema_json",
    "update_forward_refs",
    "validate",
    "model_config",
    "model_construct",
    "model_copy",
    "model_dump",
    "model_dump_json",
    "model_extra",
    "model_fields",
    "model_fields_set",
    "model_json_schema",
    "model_parametrized_name",
    "model_post_init",
    "model_rebuild",
    "model_validate",
    "model_validate_json",
    "model_validate_strings",
    "parse_raw",
    "model_computed_fields",
}


migration_version = (
    files("fused_batch")
    .joinpath("MIGRATION_VERSION")
    .read_text(encoding="utf-8")
    .strip("\n")
)
if migration_version is None:
    raise ValueError("Migration version not found.")


class FusedBaseModel(BaseModel):
    version: str = migration_version
    # _vobj__migrations: List[Any] = []

    @property
    def _api(self) -> FusedAPI:
        # Note that this does not import the FusedAPI class for circular import reasons
        # We assume that the API has already been instantiated before a model is created
        return get_api()

    def __dir__(self) -> List[str]:
        """Provide method name lookup and completion. Only provide 'public'
        methods.
        """
        # This enables autocompletion.
        normal_dir = {
            name
            for name in dir(type(self))
            if not name.startswith("_") and name not in PYDANTIC_METHODS
        }
        pydantic_fields = set(self.model_fields.keys())

        return sorted(normal_dir | pydantic_fields)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    # @model_validator(mode="before")
    # @classmethod
    # def versioning(cls, values):
    #     # Only validate versioning if the class has a _validate_version attribute.
    #     if "version" not in values or not hasattr(cls, "_validate_version"):
    #         return values
    #
    #     if len(cls._vobj__migrations) == 0:
    #         # Class has no migrations.
    #         return values
    #
    #     for version, migration, migration_cls in cls._vobj__migrations:
    #         # Migrate obj from current version through upper range.
    #         if cls == migration_cls:
    #             asset_version = values.get("version")
    #             if not asset_version:
    #                 logger.info(
    #                     "Asset has no version. Assuming it's 0.0.0.",
    #                 )
    #                 asset_version = "0.0.0"
    #             if asset_version < version:
    #                 logger.trace(f"Migrating to version {version}...")
    #                 values = migration(values)
    #                 values["version"] = version
    #                 logger.trace(f"Migration to version {version} complete.")
    #             else:
    #                 logger.trace(
    #                     f"Asset version is {values['version']}. Therefore, skipping migration to version {version}.",
    #                     cls,
    #                 )
    #
    #     return values


UserMetadataType = Optional[Dict[str, Any]]



import base64
import functools
import hashlib
import inspect
import os
import pickle
import types
from pathlib import Path
from typing import Any, Awaitable, Callable, Optional, TypeVar

from loguru import logger

BASE_PATH = "/mnt/cache"


def _serialize_code(code: types.CodeType) -> bytes:
    func_code_bytes = base64.b64encode(code.co_code)

    # Introduce variable values into hash.
    func_const_values = [
        # Embedded lambdas must be serialized, same as the overall function
        (
            _serialize_code(const)
            if isinstance(const, types.CodeType)
            else str(const).encode("utf-8")
        )
        for const in code.co_consts
    ]
    # Names must be serialized as otherwise referenced names (min vs max) would not
    # affect cache key.
    func_name_values = [name.encode("utf-8") for name in code.co_names]

    return b"".join(
        [
            func_code_bytes,
            *func_const_values,
            *func_name_values,
        ]
    )


def _serialize_function_defaults(func: Callable) -> bytes:
    parts = []

    if func.__defaults__ is not None:
        for d in func.__defaults__:
            parts.append(_hashify(d).encode("utf-8"))

    if func.__kwdefaults__ is not None:
        for key, value in func.__kwdefaults__.items():
            parts.append(f"{key}{_hashify(value)}".encode("utf-8"))

    return b"".join(parts)


def _hashify(func) -> str:
    hash_object = hashlib.sha256()
    try:
        if hasattr(func, "__fused_cached_fn"):
            return _hashify(func.__fused_cached_fn)
        elif callable(func):
            hash_object.update(_serialize_code(func.__code__))
            # Caution! The defaults and args do not go into the same part of the cache key!
            hash_object.update(_serialize_function_defaults(func))
        else:
            hash_object.update(str(func).encode("utf-8"))
        return hash_object.hexdigest()
    except Exception as e:
        logger.warning(f"Error Hashing {e}")
        return ""


def _cache(
    func: Callable,
    *args,
    reset: bool = False,
    path: str = "tmp",
    retry: bool = True,
    **kwargs,
):
    path = path.strip("/")
    # Cache in mounted drive if available & writable, else cache in /tmp
    if Path(BASE_PATH).exists() and os.access(BASE_PATH, os.W_OK):
        base_path = Path(BASE_PATH)
    else:
        base_path = Path("/tmp")

    # Cache directory
    # TODO: consider udf name in path once available from Fused global context
    path = Path(base_path) / "cached_data" / path
    path.mkdir(parents=True, exist_ok=True)

    # Pop reserved `_cache_id`kwarg
    _cache_id = kwargs.pop("_cache_id", None)

    try:
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify _cache_id
        id += _hashify(_cache_id)

        # 5. Hashify composite id
        id = _hashify(id)

        path_file = path / f"data_{id}"

        if not os.path.exists(path_file) or reset:
            with open(path_file, "wb") as f:
                data = func(*args, **kwargs)
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                return data
        else:
            try:
                with open(path_file, "rb") as f:
                    data = pickle.load(f)

                    return data
            except Exception:
                logger.debug("Cache not found, retrying.")
                if retry:
                    with open(path_file, "wb") as f:
                        data = func(*args, **kwargs)
                        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                        return data
                else:
                    return None
    except Exception as e:
        logger.info(f"Error Caching {e}")
        raise e


async def _cache_async(
    func: Callable,
    *args,
    reset: bool = False,
    path: str = "tmp",
    retry: bool = True,
    **kwargs,
):
    path = path.strip("/")
    # Cache in mounted drive if available & writable, else cache in /tmp
    if Path(BASE_PATH).exists() and os.access(BASE_PATH, os.W_OK):
        base_path = Path(BASE_PATH)
    else:
        base_path = Path("/tmp")

    # Cache directory
    # TODO: consider udf name in path once available from Fused global context
    path = Path(base_path) / "cached_data" / path
    path.mkdir(parents=True, exist_ok=True)

    # Pop reserved `_cache_id`kwarg
    _cache_id = kwargs.pop("_cache_id", None)

    try:
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify _cache_id
        id += _hashify(_cache_id)

        # 5. Hashify composite id
        id = _hashify(id)

        path_file = path / f"data_{id}"

        if not os.path.exists(path_file) or reset:
            with open(path_file, "wb") as f:
                data = await func(*args, **kwargs)
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                return data
        else:
            try:
                with open(path_file, "rb") as f:
                    data = pickle.load(f)

                    return data
            except Exception:
                logger.debug("Cache not found, retrying.")
                if retry:
                    with open(path_file, "wb") as f:
                        data = await func(*args, **kwargs)
                        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
                        return data
                else:
                    return None
    except Exception as e:
        logger.info(f"Error Caching {e}")
        raise e


def _cache_internal(func, **decorator_kwargs):
    def decorator(func):
        _path = decorator_kwargs.get("path", "tmp")

        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper_async(*args, **kwargs):
                return await _cache_async(func, *args, path=_path, **kwargs)

            wrapper_async.__fused_cached_fn = func

            return wrapper_async
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return _cache(func, *args, path=_path, **kwargs)

            wrapper.__fused_cached_fn = func

            return wrapper

    if callable(func):  # w/o args
        return decorator(func)
    else:  # w/ args
        return decorator


def cache(
    func: Optional[Callable[..., Any]] = None, **kwargs: Any
) -> Callable[..., Any]:
    """Decorator to cache the return value of a function.

    This function serves as a decorator that can be applied to any function
    to cache its return values. The cache behavior can be customized through
    keyword arguments.

    Args:
        func (Callable, optional): The function to be decorated. If None, this
            returns a partial decorator with the passed keyword arguments.
        **kwargs: Arbitrary keyword arguments that are passed to the internal
            caching mechanism. These could specify cache size, expiration time,
            and other cache-related settings.

    Returns:
        Callable: A decorator that, when applied to a function, caches its
        return values according to the specified keyword arguments.

    Examples:

        Use the `@cache` decorator to cache the return value of a function in a custom path.

        ```py
        @cache(path="/tmp/custom_path/")
        def expensive_function():
            # Function implementation goes here
            return result
        ```

        If the output of a cached function changes, for example if remote data is modified,
        it can be reset by running the function with the `reset` keyword argument. Afterward,
        the argument can be cleared.

        ```py
        @cache(path="/tmp/custom_path/", reset=True)
        def expensive_function():
            # Function implementation goes here
            return result
        ```
    """
    return _cache_internal(func=func, **kwargs)


T = TypeVar("T")


def cache_call(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Directly calls a function with caching.

    This function directly calls the provided function with the given arguments
    and keyword arguments, caching its return value. The cache used depends on
    the implementation of the `_cache` function.

    Args:
        func (Callable): The function to call and cache its result.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The cached return value of the function.

    Raises:
        Exception: Propagates any exception raised by the function being called
        or the caching mechanism.
    """
    return _cache(func, *args, **kwargs)


async def cache_call_async(
    func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> T:
    """Asynchronously calls a function with caching.

    Similar to `cache_call`, but for asynchronous functions. This function
    awaits the provided async function, caches its return value, and then
    returns it. The specifics of the caching mechanism depend on the
    implementation of `_cache_async`.

    Args:
        func (Callable): The asynchronous function to call and cache its result.
        *args: Positional arguments to pass to the async function.
        **kwargs: Keyword arguments to pass to the async function.

    Returns:
        The cached return value of the async function.

    Raises:
        Exception: Propagates any exception raised by the async function being
        called or the caching mechanism.

    Examples:
        async def fetch_data(param):
            # Async function implementation goes here
            return data

        # Usage
        result = await cache_call_async(fetch_data, 'example_param')
    """
    return await _cache_async(func, *args, **kwargs)



import warnings
from pathlib import Path
from typing import Any, BinaryIO, Iterable, Literal, Optional, Union, overload

from fused_batch.models.api.job import (
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    JoinType,
    MapJobStepConfig,
    UdfJobStepConfig,
)
from fused_batch.models.internal import DatasetOutputBase
from fused_batch.warnings import FusedUdfWarning

from ..coerce_dataset import CoerceableToDatasetInput, _object_to_dataset_input
from .base_udf import UdfType
from .udf import GeoPandasUdfV2


class GeoPandasUdfV2Callable(GeoPandasUdfV2):
    type: Literal[UdfType.GEOPANDAS_V2] = UdfType.GEOPANDAS_V2
    """This class is returned from `@fused_batch.udf` and represents
    a UDF that can be instantiated into a job."""

    def to_file(self, where: Union[str, Path, BinaryIO], *, overwrite: bool = False):
        """Write the UDF to disk or the specified file-like object.

        The UDF will be written as a Zip file.

        Args:
            where: A path to a file or a file-like object.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="zip", overwrite=overwrite)

    def to_directory(self, where: Union[str, Path], *, overwrite: bool = False):
        """Write the UDF to disk as a directory (folder).

        Args:
            where: A path to a directory.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="local", overwrite=overwrite)

    def to_gist(self, where: Optional[str] = None, *, overwrite: bool = False):
        """Write the UDF to Github as a Gist.

        Args:
            gist_id: Optionally, a Gist ID to overwrite.

        Keyword Args:
            overwrite: If true, overwriting is allowed.
        """
        job = self()
        job.export(where, how="gist", overwrite=overwrite)

    # This is a subclass of GeoPandasUdfV2 so that the job classes can reference
    # GeoPandasUdfV2 without issues. This class is then installed over the
    # GeoPandasUdfV2 type code so that loaded objects get the __call__ methods.

    # List of data input is passed - run that
    @overload
    def __call__(
        self,
        dataset: None = None,
        right: None = None,
        *,
        arg_list: Iterable[Any],
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_is_singlefile: Optional[bool] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        **kwargs
    ) -> UdfJobStepConfig:
        ...

    # Nothing is passed - run the UDF once
    @overload
    def __call__(
        self,
        dataset: None = None,
        right: None = None,
        *,
        arg_list: None = None,
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_is_singlefile: Optional[bool] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        **kwargs
    ) -> UdfJobStepConfig:
        ...

    @overload
    def __call__(
        self,
        dataset: CoerceableToDatasetInput,
        right: None = None,
        *,
        arg_list: None = None,
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_is_singlefile: Optional[bool] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        **kwargs
    ) -> MapJobStepConfig:
        ...

    @overload
    def __call__(
        self,
        dataset: CoerceableToDatasetInput,
        right: str,
        *,
        arg_list: None = None,
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        join_is_singlefile: Literal[True],
        **kwargs
    ) -> JoinSinglefileJobStepConfig:
        ...

    # None for join_is_singlefile will get typed as JoinJobStepConfig,
    # which is probably fine. (single file is expected to be rare/unusual.)
    # Specify join_is_singlefile explicitly if that's an issue.
    @overload
    def __call__(
        self,
        dataset: CoerceableToDatasetInput,
        right: CoerceableToDatasetInput,
        *,
        arg_list: None = None,
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_is_singlefile: Optional[Literal[False]] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        **kwargs
    ) -> JoinJobStepConfig:
        ...

    def __call__(
        self,
        dataset: Optional[CoerceableToDatasetInput] = None,
        right: Optional[CoerceableToDatasetInput] = None,
        *,
        arg_list: Optional[Iterable[Any]] = None,
        output_table: Optional[str] = None,
        buffer_distance: Optional[float] = None,
        join_is_singlefile: Optional[bool] = None,
        join_how: Union[JoinType, Literal["left", "inner"]] = JoinType.INNER,
        **kwargs
    ) -> Union[
        UdfJobStepConfig,
        MapJobStepConfig,
        JoinJobStepConfig,
        JoinSinglefileJobStepConfig,
    ]:
        """Create a job from this UDF.

        Args:
            dataset: The dataset to run the UDF on.
            right: The right dataset to join with. If None, no join is performed and only `dataset` is processed. If this is a URL to a single Parquet file, a singlefile join is created. Defaults to None.
            arg_list: A list of records to pass in to the UDF as input. This option is mutually exclusive with `dataset` and `right`.
            output_table: The name of the table to write output columns on `dataset`.
            buffer_distance: For join jobs, the buffer around `dataset` partitions to perform.
            join_is_singlefile: If not None, whether a join operation should be performed in `singlefile` mode, i.e. all partitions join with a single Parquet file. Defaults to None, which indicates autodetect.
        """

        any_datasets_passed = dataset is not None or right is not None
        any_data_passed = arg_list is not None
        if any_datasets_passed and any_data_passed:
            warnings.warn(
                "`dataset` and `arg_list` are mutually exclusive options, `dataset` will be ignored.",
                FusedUdfWarning,
            )

        passed_dataset = dataset
        dataset = (
            _object_to_dataset_input(dataset, allow_none=True)
            if any_datasets_passed
            else None
        )
        if dataset is None:
            # Detect that this is actually some sort of regular data input, not a dataset.
            any_datasets_passed = False

        with_params = self.model_copy()
        # TODO: Consider using with_parameters here, and validating that "context" and other reserved parameter names are not being passed.
        new_parameters = {**kwargs}
        if new_parameters:
            with_params.parameters = new_parameters

        project_url = dataset._project_url if hasattr(dataset, "_project_url") else None

        output = DatasetOutputBase.from_str(output_table, project_url=project_url)

        if any_data_passed or not any_datasets_passed:
            if arg_list is not None and not len(arg_list):
                warnings.warn(
                    "An empty `arg_list` was passed in, no calls to the UDF will be made.",
                    FusedUdfWarning,
                )
            if passed_dataset is not None:
                arg_list = (
                    [passed_dataset, *arg_list]
                    if arg_list is not None
                    else [passed_dataset]
                )

            return UdfJobStepConfig(
                udf=with_params,
                input=arg_list,
                output=output,
            )
        elif right is None:
            assert dataset is not None
            return MapJobStepConfig(
                input=dataset,
                udf=with_params,
                output=output,
            )
        elif (
            isinstance(right, str) and right.endswith(".parquet")
        ) or join_is_singlefile is True:
            assert dataset is not None
            return JoinSinglefileJobStepConfig(
                input_left=dataset,
                input_right=right,
                udf=with_params,
                output=output,
                buffer_distance=buffer_distance,
            )
        else:
            assert dataset is not None
            right = _object_to_dataset_input(right)
            return JoinJobStepConfig(
                input_left=dataset,
                input_right=right,
                udf=with_params,
                output=output,
                buffer_distance=buffer_distance,
                how=join_how,
            )



import urllib.parse
from datetime import datetime
from typing import Any, Dict, Optional, Union

from fused_batch._formatter.formatter_udf_access_token import (
    fused_udf_access_token_list_repr,
    fused_udf_access_token_repr,
)
from fused_batch.models.base import FusedBaseModel

from .._inplace import _maybe_inplace


def _make_dtype_query_params(
    *, dtype_out_vector: Optional[str] = None, dtype_out_raster: Optional[str] = None
):
    ret = urllib.parse.urlencode(
        {
            **({"dtype_out_vector": dtype_out_vector} if dtype_out_vector else {}),
            **({"dtype_out_raster": dtype_out_raster} if dtype_out_raster else {}),
        }
    )
    if ret:
        return f"?{ret}"
    return ret


class UdfAccessToken(FusedBaseModel):
    token: str
    udf_email: Optional[str] = None
    udf_slug: Optional[str] = None
    udf_id: Optional[str] = None
    enabled: bool
    owning_user_id: str
    client_id: Optional[str]
    cache: Optional[bool] = None
    metadata_json: Optional[Dict[str, Any]]
    last_updated: datetime

    def update(
        self,
        client_id: Optional[str] = None,
        cache: Optional[bool] = None,
        metadata_json: Optional[Dict[str, Any]] = None,
        enabled: Optional[bool] = None,
        inplace: bool = False,
    ) -> "UdfAccessToken":
        ret = _maybe_inplace(self, inplace)
        return ret.model_copy(
            update=self._api.update_udf_access_token(
                token=self.token,
                client_id=client_id,
                cache=cache,
                metadata_json=metadata_json,
                enabled=enabled,
            ).model_dump()
        )

    def refresh(self, inplace: bool = False) -> "UdfAccessToken":
        ret = _maybe_inplace(self, inplace)
        return ret.model_copy(
            update=self._api.get_udf_access_token(self.token).model_dump()
        )

    def delete(self) -> "UdfAccessToken":
        return self._api.delete_udf_access_token(self.token)

    def get_file_url(
        self,
        *,
        dtype_out_vector: Optional[str] = None,
        dtype_out_raster: Optional[str] = None,
    ) -> str:
        query_params = _make_dtype_query_params(
            dtype_out_vector=dtype_out_vector, dtype_out_raster=dtype_out_raster
        )
        return (
            f"{self._api.base_url}/realtime-shared/{self.token}/run/file{query_params}"
        )

    def get_tile_url(
        self,
        *,
        x: Union[int, str, None] = None,
        y: Union[int, str, None] = None,
        z: Union[int, str, None] = None,
        dtype_out_vector: Optional[str] = None,
        dtype_out_raster: Optional[str] = None,
    ) -> str:
        query_params = _make_dtype_query_params(
            dtype_out_vector=dtype_out_vector, dtype_out_raster=dtype_out_raster
        )
        if x is None and y is None and z is None:
            x = "{x}"
            y = "{y}"
            z = "{z}"
        elif x is None or y is None or z is None:
            raise ValueError("All of x, y, and z must be specified")
        return f"{self._api.base_url}/realtime-shared/{self.token}/run/tiles/{z}/{x}/{y}{query_params}"

    def _repr_html_(self) -> str:
        return fused_udf_access_token_repr(self)


class UdfAccessTokenList(list):
    def _repr_html_(self) -> str:
        return fused_udf_access_token_list_repr(self)



from typing import Optional


def get_secret_from_server(key: str, client_id: Optional[str] = None):
    from fused.api.api import FusedAPI

    api = FusedAPI()
    return api.get_secret_value(key=key, client_id=client_id)


def list_secrets_from_server(client_id: Optional[str] = None):
    from fused.api.api import FusedAPI

    api = FusedAPI()
    return api.list_secrets(client_id=client_id)


def set_secret_on_server(key: str, value: str, client_id: Optional[str] = None):
    from fused.api.api import FusedAPI

    api = FusedAPI()
    return api.set_secret_value(key=key, value=value, client_id=client_id)


def delete_secret_on_server(key: str, client_id: Optional[str] = None):
    from fused.api.api import FusedAPI

    api = FusedAPI()
    return api.delete_secret_value(key=key, client_id=client_id)
